#include "filesys.h"
#include <stdio.h>

int dir_listing_co(struct dir_listing *self, const char *path)
{
#ifdef WIN32
	TCHAR dirspec[MAX_PATH];

	self->end_of_listing = 0;

	if(strlen(path) > (MAX_PATH - 3)){
		return -1;
	}
	sprintf(dirspec, "%s\\*", path);

	self->handle = FindFirstFile(dirspec, &self->find_data);
	if(self->handle == INVALID_HANDLE_VALUE){
		return -1;
	}
	return 0;
#elif LINUX
	self->dir = opendir(path);
	if(self->dir == NULL) return -1;
	return 0;
#endif
}
void dir_listing_de(struct dir_listing *self)
{
#ifdef WIN32
	FindClose(self->handle);
#elif LINUX
	if(self->dir != NULL) closedir(self->dir);
#endif
}
int dir_listing_get_next(struct dir_listing *self, int *type, char *name, unsigned int maxlen)
{
#ifdef WIN32
	if(self->end_of_listing) return 1;
	snprintf(name, maxlen, self->find_data.cFileName);
	if(self->find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) *type = FS_DIR;
	else *type = FS_FILE;
	if(FindNextFile(self->handle, &self->find_data) == 0){
		self->end_of_listing = 1;
	}
	return 0;
#elif LINUX
	struct dirent *dp = readdir(self->dir);
	if(dp == NULL) return 1;
	snprintf(name, maxlen, dp->d_name);
	if(dp->d_type == DT_REG) *type = FS_FILE;
	else if(dp->d_type == DT_DIR) *type = FS_DIR;
	else *type = FS_OTHER;
	return 0;
#endif
}


