#include <math.h>
#include <assert.h>
#include "game.h"
#include "common.h"
#include "gfx.h"

int game_co(struct game *self, SDL_Surface *screen)
{
	unsigned int i;
	assert(screen->w == 640 && screen->h == 480);
	self->screen = screen;
	if(imglib_co(&self->imglib)) return -1;
	for(i=0; i<UNIT_LIST_SIZE; i++)
	{
		if(unit_co(&self->units[i], i)==-1) return -1;
		//fprintf(stderr, "succesfully ran unit_co(&self->units[%i]).\n", i);
		assert(self->units[i].type_id == -1);
	}
	map_co(&self->map);
	self->view.w = screen->w;
	self->view.h = screen->h;
	self->view.x = 0;
	self->view.y = 0;
	self->selected_units_count = 0;
	self->settings.reach_target_always_when_other_reached_it_too = 0;
	return 0;
}

void game_de(struct game *self)
{
	unsigned int i;
	map_de(&self->map);
	for(i=0; i<UNIT_LIST_SIZE; i++)
	{
		unit_de(&self->units[i]);
	}
	imglib_de(&self->imglib);
}

int game_load(struct game *self)
{
	imglib_set_format(&self->imglib, self->screen->format, self->screen->flags, SDL_MapRGB(self->screen->format, 0,255,0));
	//imglib_set_format(&self->imglib, self->screen->format, self->screen->flags, 2);
	if(imglib_load(&self->imglib, "./images")){
		fprintf(stderr, "imglib_load failed\n");
		return -1;
	}
	unit_specs_initialize(&self->imglib);
	
	map_load_string(&self->map, testmap, &self->imglib);

	return 0;
}

int game_run(struct game *self)
{
	fprintf(stderr, "game_run\n");

	self->screen_needs_complete_update = 1;
	Uint32 ms_per_frame = 1000/15;
	//Uint32 ms_per_frame = 1000;
	Uint32 oldticks = 0;
	unsigned int i;
	unsigned int framecount=0;
	unsigned int scrollspeed = 200/15;
	//unsigned int scrollspeed = 1;
	int mouse_x=0, mouse_y=0;
	unsigned int cursor_edge_scroll_width = 10;

	//ts = imglib_get_image(&self->imglib, "test.bmp");
	
	//assert(unit_load(&self->units[0], 0)==0);
	//assert(unit_set_tpos(&self->units[0], 2, 2, -1)==0);
	/*for(i=0; i<98; i++){
		assert(unit_load(&self->units[i], 0)==0);
		assert(unit_set_tpos(&self->units[i], 1+i/7, 1+i%7, -1)==0);
	}*/
	/*for(i=0; i<100; i++){
		assert(unit_load(&self->units[i+100], 1)==0);
		assert(unit_set_tpos(&self->units[i+100], 1+i/6, 8+i%6, -1)==0);
	}*/
	assert(unit_load(&self->units[200], 0)==0);
	assert(unit_set_tpos(&self->units[200], 2, 7, -1)==0);
	assert(unit_load(&self->units[201], 0)==0);
	assert(unit_set_tpos(&self->units[201], 14, 7, -1)==0);
	assert(unit_load(&self->units[202], 1)==0);
	assert(unit_set_tpos(&self->units[202], 15, 7, -1)==0);
	for(i=0; i<20; i++){
		assert(unit_load(&self->units[i], 0)==0);
		assert(unit_set_tpos(&self->units[i], 1+i/5, 1+i%5, -1)==0);
		game_selection_add(self, i);
	}
	/*assert(unit_set_tpos(&self->units[17], 16, 5, -1)==0);
	assert(unit_set_tpos(&self->units[18], 16, 6, -1)==0);
	assert(unit_set_tpos(&self->units[14], 16, 7, -1)==0);
	assert(unit_set_tpos(&self->units[12], 17, 5, -1)==0);
	assert(unit_set_tpos(&self->units[13], 17, 6, -1)==0);
	assert(unit_set_tpos(&self->units[19], 17, 7, -1)==0);
	game_selection_add(self, 17);
	game_selection_add(self, 18);
	game_selection_add(self, 14);
	game_selection_add(self, 12);
	game_selection_add(self, 13);
	game_selection_add(self, 19);*/

	for(;;framecount++)
	{
		SDL_Event event;
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
			case SDL_KEYDOWN:
				break;
			case SDL_KEYUP:
				switch(event.key.keysym.sym)
				{
				case SDLK_ESCAPE:
					return 0;
				default:
					break;
				}
				break;
			case SDL_MOUSEMOTION:
				mouse_x = event.motion.x;
				mouse_y = event.motion.y;
				break;
			case SDL_MOUSEBUTTONDOWN:
				if(event.button.button == SDL_BUTTON_LEFT)
				{
					int tx = (event.button.x+self->view.x)/MAP_TILE_W;
					int ty = (event.button.y+self->view.y)/MAP_TILE_H;
					int ui = game_select_unit_at(self, event.button.x+self->view.x, event.button.y+self->view.y);
					if(ui==-1 && self->selected_units_count == 0){
						fprintf(stderr, "clicked on map (%i,%i) (and no units are selected)\n", tx, ty);
					}
					else if(ui==-1){
						fprintf(stderr, "clicked on map (%i,%i) (and some units are selected)\n", tx, ty);
						assert(self->selected_units_count <= SELECTED_UNITS_SIZE);
						unsigned int i;
						uint8_t all_units_do_nothing = 1;
						for(i=0; i<self->selected_units_count; i++){
							unsigned int ui = self->selected_units[i];
							assert(ui < UNIT_LIST_SIZE);
							if(unit_click_map(&self->units[ui], tx, ty, self->selected_units_count) != -1)
								all_units_do_nothing = 0;
						}
						if(all_units_do_nothing){
							game_selection_clear(self);
						}
					}
					else{
						fprintf(stderr, "clicked on unit %i\n", ui);
						game_selection_add(self, ui);
					}
				}
				else if(event.button.button == SDL_BUTTON_RIGHT)
				{
					game_selection_clear(self);
				}
				break;
			case SDL_QUIT:
				return 0;
			default:
				break;
			}
		}
		
		int oldvx = self->view.x;
		int oldvy = self->view.y;
		if(mouse_x < cursor_edge_scroll_width){
			if(self->view.x > scrollspeed)
				self->view.x -= scrollspeed;
			else
				self->view.x = 0;
		}
		if(mouse_y < cursor_edge_scroll_width){
			if(self->view.y > scrollspeed)
				self->view.y -= scrollspeed;
			else
				self->view.y = 0;
		}
		if(mouse_x >= self->screen->w - cursor_edge_scroll_width){
			if(self->view.x + self->view.w < self->map.w * MAP_TILE_W - scrollspeed)
				self->view.x += scrollspeed;
			else
				self->view.x = self->map.w * MAP_TILE_W - self->view.w;
		}
		if(mouse_y >= self->screen->h - cursor_edge_scroll_width){
			if(self->view.y + self->view.h < self->map.h * MAP_TILE_H - scrollspeed)
				self->view.y += scrollspeed;
			else
				self->view.y = self->map.h * MAP_TILE_H - self->view.h;
		}
		if(oldvx != self->view.x || oldvy != self->view.y)
			self->screen_needs_complete_update = 1;

		//fprintf(stderr, "view.x=%i, view.y=%i\n", self->view.x, self->view.y);

		//game action here

		for(i=0; i<UNIT_LIST_SIZE; i++)
		{
			if(self->units[i].type_id == -1) continue;
			unit_action(&self->units[i], self);
		}


		/*if(tempcount < 1.0) tempcount+=0.01;
		else tempcount = 0.;
		assert(unit_set_tpos(&self->units[0], (int)(5.+5.*cos(tempcount*2*PI)), (int)(5.+5.*sin(tempcount*2*PI)), -1)==0);*/

		//end of game action
		
		//fprintf(stderr, "drawing game\n");
		if(game_draw(self) == -1){
			fprintf(stderr, "game_draw failed\n");
			return -1;
		}

		Uint32 newticks = SDL_GetTicks();
		if(newticks < oldticks){
			oldticks = newticks;
			continue;
		}
		if(newticks >= oldticks + ms_per_frame){
			oldticks = newticks;
			//fprintf(stderr, "frame %i: can't keep up with wanted frame time\n", framecount);
			continue;
		}
		Uint32 waittime = ms_per_frame - (newticks - oldticks);
		//fprintf(stderr, "waittime=%i\n", waittime);
		oldticks = newticks;
		SDL_Delay(waittime);
	}
	return 0;
}

int game_draw(struct game *self)
{
	unsigned int i;

	if(SDL_MUSTLOCK(self->screen))
		if(SDL_LockSurface(self->screen)==-1)
			return -1;

	/*int j;
	for(j=0;j<480;j++){
		uint8_t *p = self->screen->pixels + j*self->screen->pitch;
		int i;
		for(i=0; i<640; i++){
			*p = 0x50;
			p++;
		}
	}*/
	
	if(self->screen_needs_complete_update){
		map_draw(&self->map, self->screen, self->view.x, self->view.y, 0, 0, self->screen->w, self->screen->h);
	}
	else{
		for(i=0; i<UNIT_LIST_SIZE; i++)
		{
			if(self->units[i].type_id == -1) continue;
			int num;
			unsigned int j;
			struct unit *unit = &self->units[i];
			
			if(unit->graphics_changed == 0) continue;
			
			if(self->screen_needs_complete_update == 0){
				SDL_Rect rects[UNIT_MAX_DRAW_IMAGE_COUNT];
				num = unit_get_last_draw_rects(unit, rects, UNIT_MAX_DRAW_IMAGE_COUNT);
				if(num==-1){
					fprintf(stderr, "error in unit_get_last_draw_rects\n");
				}
				for(j=0; j<num; j++){
					map_draw(&self->map, self->screen, rects[j].x, rects[j].y, rects[j].x-self->view.x, rects[j].y-self->view.y, rects[j].w, rects[j].h);
				}
			}
			if(unit->selected || unit->was_selected){
				SDL_Rect area, areascreen;
				if(unit_get_select_area(&self->units[i], &area)==-1){
					fprintf(stderr, "unit_get_select_area(&self->units[%i], &area) failed\n", i);
					area.w = 0;
					area.h = 0;
				}
				else{
					areascreen.x = area.x - self->view.x;
					areascreen.y = area.y - self->view.y;
					areascreen.w = area.w;
					areascreen.h = area.h;
				}
				map_draw(&self->map, self->screen, area.x, area.y, areascreen.x, areascreen.y, area.w, area.h);
			}
		}
	}
	
	for(i=0; i<UNIT_LIST_SIZE; i++)
	{
		if(self->units[i].type_id == -1) continue;
		int num;
		unsigned int j;
		struct unit *unit = &self->units[i];
		
		/*
			If the current unit doesn't need to be updated, and the whole
			screen doesn't need to be updated (which would mean this too would
			have to be.)
			//
			That doesn't work, because some units need to get updated while
			their graphics_changed == 0, because of certain diagonal moves.
			TODO:
			check those moves somewhere and set graphics_changeds accordingly
		*/
		//if(self->screen_needs_complete_update == 0 && unit->graphics_changed == 0) continue;

		struct draw_image_info info[UNIT_MAX_DRAW_IMAGE_COUNT];
		num = unit_get_draw_info(unit, info, UNIT_MAX_DRAW_IMAGE_COUNT);
		if(num==-1){
			fprintf(stderr, "error in unit_get_draw_info\n");
		}
		for(j=0; j<num; j++){
			SDL_Surface *ts = info[j].s;
			SDL_Rect dst;
			dst.x = info[j].px - self->view.x;
			dst.y = info[j].py - self->view.y;
			SDL_BlitSurface(ts, NULL, self->screen, &dst);
		}
		if(unit->selected){
			SDL_Rect area, areascreen;
			if(unit_get_select_area(&self->units[i], &area)==-1){
				fprintf(stderr, "unit_get_select_area(&self->units[%i], &area) failed\n", i);
				area.w = 0;
				area.h = 0;
			}
			else{
				areascreen.x = area.x - self->view.x;
				areascreen.y = area.y - self->view.y;
				areascreen.w = area.w;
				areascreen.h = area.h;
			}
			gfx_draw_selection_cursor(self->screen, areascreen, SDL_MapRGB(self->screen->format, 255,255,255));
		}
		unit->graphics_changed = 0;
		unit->was_selected = 0;
	}
	
	/*int a = SDL_GetTicks();
	int b;
	for(b=0; b<100000; b++){
		gfx_draw_selection_cursor(self->screen, 32, 32, 32, 32, SDL_MapRGB(self->screen->format, 255,255,255));
	}
	printf("%fus\n", (double)(SDL_GetTicks()-a)/100);*/

	SDL_UpdateRect(self->screen, 0, 0, 640, 480);

	if(self->screen_needs_complete_update){
		self->screen_needs_complete_update = 0;
	}

	if(SDL_MUSTLOCK(self->screen))
		SDL_UnlockSurface(self->screen);
	
	return 0;
}

int game_get_tile_info(struct game *self, struct tile_info *info, int tx, int ty)
{
	if(tx<0 || ty<0 || tx>=self->map.w || ty>=self->map.h){
		return -1;
		//info->mapflags = 0;
	}
	else{
		info->mapflags = self->map.tiles[ty*self->map.w+tx].flags;
	}
	//fprintf(stderr, " game_get_tile_info(%i,%i): mapflags=0x%x\n", tx, ty, info->mapflags);
	unsigned int i, j=0;
	for(i=0; i<UNIT_LIST_SIZE; i++)
	{
		//fprintf(stderr, "game_get_tile_info: checking unit %i\n", i);
		if(self->units[i].type_id == -1) continue;
		struct tpos ts[UNIT_MAX_BLOCKED_TILES];
		int numtiles = unit_get_blocked_tiles(&self->units[i], ts);
		assert(numtiles != -1);
		unsigned int k;
		//fprintf(stderr, "game_get_tile_info: unit %i says to be on: ", i);
		for(k=0; k<numtiles; k++){
			//fprintf(stderr, "(%i,%i), ", ts[k].x, ts[k].y);
			if(ts[k].x==tx && ts[k].y==ty){
				//the unit is on this tile
				if(j>=MAX_UNITS_PER_TILE){
					fprintf(stderr, "game_get_tile_info: tile %i,%i has more than MAX_UNITS_PER_TILE=%i units\n", tx, ty, MAX_UNITS_PER_TILE);
					assert(0); //DEBUG
					break;
				}
				info->unit_i[j] = i;
				j++;
				//doesn't matter if the same unit says to be more than once on
				//the tile
				break;
			}
		}
		//fprintf(stderr, "\n");
	}
	info->unit_count = j;
	return 0;
}

int game_select_unit_at(struct game *self, int px, int py)
{
	unsigned int i;
	for(i=0; i<UNIT_LIST_SIZE; i++)
	{
		if(self->units[i].type_id == -1) continue;
		SDL_Rect area;
		if(unit_get_select_area(&self->units[i], &area)==-1){
			fprintf(stderr, "game_select_unit_at(): px=%i,py=%i: unit_get_select_area(&self->units[%i], &area) failed\n", px, py, i);
			continue;
		}
		if(px >= area.x && px < area.x + area.w && py >= area.y && py < area.y + area.h) return i;
	}
	return -1;
}
int game_selection_add(struct game *self, unsigned int i)
{
	assert(self->selected_units_count <= SELECTED_UNITS_SIZE);
	assert(i>=0 && i<UNIT_LIST_SIZE);
	if(self->units[i].type_id == -1) return -1;
	if(self->selected_units_count >= SELECTED_UNITS_SIZE) return -1;
	self->selected_units[self->selected_units_count] = i;
	self->selected_units_count++;
	self->units[i].selected = 1;
	self->units[i].graphics_changed = 1;
	return 0;
}
void game_selection_clear(struct game *self)
{
	assert(self->selected_units_count <= SELECTED_UNITS_SIZE);
	unsigned int i;
	for(i=0; i<self->selected_units_count; i++){
		unsigned int ui = self->selected_units[i];
		assert(ui < UNIT_LIST_SIZE);
		self->units[ui].selected = 0;
		self->units[ui].was_selected = 1;
		self->units[ui].graphics_changed = 1;
	}
	self->selected_units_count = 0;
}

int8_t game_path_get_blocked(struct game *self, int tx, int ty, blockedfunc_t blockedfunc)
{
	struct tile_info ti;
	if(tx<0 || ty<0 || tx>=self->map.w || ty>=self->map.h){
		ti.mapflags = 0;
		ti.unit_count = 0;
		return (*blockedfunc)(&ti);
	}
	if(self->map.tiles[ty*self->map.w+tx].blocked == -2){
		if(game_get_tile_info(self, &ti, tx, ty) == -1){
			self->map.tiles[ty*self->map.w+tx].blocked = -1;
			fprintf(stderr, "game_path_get_blocked(): warning: game_get_tile_info(%i,%i) failed\n", tx, ty);
			assert(0);
		}
		else{
			self->map.tiles[ty*self->map.w+tx].blocked = (*blockedfunc)(&ti);
			//fprintf(stderr, "game_path_get_blocked(%i,%i): blocked=%i (ti.mapflags=0x%x)\n", tx, ty, self->map.tiles[ty*self->map.w+tx].blocked, ti.mapflags);
		}
	}
	return self->map.tiles[ty*self->map.w+tx].blocked;
}

float path_heuristic(int x1, int y1, int x2, int y2)
{
	return distance(x1, y1, x2, y2);
}

struct tpos * game_find_path(struct game *self, int from_x, int from_y, int to_x, int to_y, blockedfunc_t blockedfunc, struct tpos_linked *additional_blocked)
	{
		//fprintf(stderr, "game_find_path(): %i,%i -> %i,%i\n", from_x, from_y, to_x, to_y);
		struct tpos *path = NULL;
		unsigned int x, y;
		struct tile_info ti;
		struct maptile *tile = self->map.tiles;
		struct tpos_linked *t, *t2, *t3;
		//the set of nodes already evaluated. not sorted.
		struct tpos_linked *closedset_first = NULL;
		//the set of tentative nodes to be evaluated. sorted by f value, smallest first
		struct tpos_linked *openset_first = NULL;
		struct tpos_linked *initial_node = (struct tpos_linked*)malloc(sizeof(struct tpos_linked));
		if(initial_node == NULL) goto path_fail;
		initial_node->x = from_x;
		initial_node->y = from_y;
		initial_node->next = NULL;
		openset_first = initial_node;
	self->map.tiles[from_y*self->map.w+from_x].g = 0.;
	self->map.tiles[from_y*self->map.w+from_x].h = path_heuristic(from_x, from_y, to_x, to_y);
	self->map.tiles[from_y*self->map.w+from_x].f = self->map.tiles[from_y*self->map.w+from_x].h;

	for(y=0; y<self->map.h; y++){
		for(x=0; x<self->map.w; x++){
			struct tpos_linked *t = additional_blocked;
			while(t!=NULL){
				if(t->x == x && t->y == y){
					tile->blocked = -1;
					break;
				}
				t = t->next;
			}
			if(t == NULL) tile->blocked = -2;
			tile->from = -1;
			tile++;
		}
	}
	
	assert(game_get_tile_info(self, &ti, from_x, from_y)!=-1);
	/*if((*blockedfunc)(&ti) == -1){
		fprintf(stderr, "game_find_path(): warning: unit is on a tile on which it can't move\n");
	}*/
	
	unsigned int path_found = 0;
	while(openset_first != NULL){
	
		assert(openset_first != closedset_first);
		if(openset_first) assert(openset_first != openset_first->next);
		if(closedset_first) assert(closedset_first != closedset_first->next);

		t = openset_first;
		if(t->x == to_x && t->y == to_y){
			path_found = 1;
			break;
		}

		//move t from openset to closedset
		openset_first = t->next;
		t->next = closedset_first;
		closedset_first = t;

		//get the neighboring nodes

		struct tpos neighbors[8];
		unsigned int neighbor_count = 0;
		int8_t left = game_path_get_blocked(self, t->x-1, t->y, blockedfunc);
		int8_t right = game_path_get_blocked(self, t->x+1, t->y, blockedfunc);
		int8_t top = game_path_get_blocked(self, t->x, t->y-1, blockedfunc);
		int8_t bottom = game_path_get_blocked(self, t->x, t->y+1, blockedfunc);
		if(left!=-1){
			neighbors[neighbor_count].x = t->x-1;
			neighbors[neighbor_count].y = t->y;
			neighbor_count++;
		}
		if(right!=-1){
			neighbors[neighbor_count].x = t->x+1;
			neighbors[neighbor_count].y = t->y;
			neighbor_count++;
		}
		if(top!=-1){
			neighbors[neighbor_count].x = t->x;
			neighbors[neighbor_count].y = t->y-1;
			neighbor_count++;
		}
		if(bottom!=-1){
			neighbors[neighbor_count].x = t->x;
			neighbors[neighbor_count].y = t->y+1;
			neighbor_count++;
		}
		//check if can go nicely to corner neighbors and add them if appropriate
		int8_t tl = game_path_get_blocked(self, t->x-1, t->y-1, blockedfunc);
		int8_t tr = game_path_get_blocked(self, t->x+1, t->y-1, blockedfunc);
		int8_t bl = game_path_get_blocked(self, t->x-1, t->y+1, blockedfunc);
		int8_t br = game_path_get_blocked(self, t->x+1, t->y+1, blockedfunc);
		if(left!=-1 && top!=-1 && tl!=-1){
			neighbors[neighbor_count].x = t->x-1;
			neighbors[neighbor_count].y = t->y-1;
			neighbor_count++;
		}
		if(left!=-1 && bottom!=-1 && bl!=-1){
			neighbors[neighbor_count].x = t->x-1;
			neighbors[neighbor_count].y = t->y+1;
			neighbor_count++;
		}
		if(right!=-1 && top!=-1 && tr!=-1){
			neighbors[neighbor_count].x = t->x+1;
			neighbors[neighbor_count].y = t->y-1;
			neighbor_count++;
		}
		if(right!=-1 && bottom!=-1 && br!=-1){
			neighbors[neighbor_count].x = t->x+1;
			neighbors[neighbor_count].y = t->y+1;
			neighbor_count++;
		}
		unsigned int i;
		for(i=0; i<neighbor_count; i++){
			struct tpos *n = &neighbors[i];
			//fprintf(stderr, "n = &neighbors[%i], n->x=%i, n->y=%i\n", i, n->x, n->y);
			//skip if n is in closedset
			t2 = closedset_first;
			while(t2 != NULL){
				if(n->x == t2->x && n->y == t2->y) break;
				t2 = t2->next;
			}
			if(t2 != NULL) continue;
			float d = distance(t->x, t->y, n->x, n->y);
			int8_t b;
			if(t->x == from_x && t->y == from_y) b = 0;
			else b = game_path_get_blocked(self, t->x, t->y, blockedfunc);
			assert(b!=-1);
			d += (float)b/10.;
			float tentative_g = self->map.tiles[t->y*self->map.w+t->x].g + d;
			//fprintf(stderr, "node n=(%i,%i): g(t)=%f, d(t,n)=%f, g(n)=%f\n", n->x, n->y, self->map.tiles[t->y*self->map.w+t->x].g, d, tentative_g);
			int tentative_h = path_heuristic(n->x, n->y, to_x, to_y);
			int tentative_f = tentative_g + tentative_h;
			//check if n is in openset (is if t2 != NULL)
			//if it is, t2 is the occurrence of n in openset
			struct tpos_linked *nl = openset_first;
			while(nl != NULL){
				if(n->x == nl->x && n->y == nl->y) break;
				nl = nl->next;
			}
			int tentative_is_better = 0;
			//if n isn't in openset
			if(nl == NULL){
				nl = (struct tpos_linked*)malloc(sizeof(struct tpos_linked));
				if(nl == NULL) goto path_fail;
				nl->x = n->x;
				nl->y = n->y;

				//fprintf(stderr, "got node (%i,%i)\n", nl->x, nl->y);
				
				//add nl to openset
				
				//search for the right place for nl in openset
				t2 = openset_first;
				t3 = NULL;
				while(t2 != NULL){
					if(self->map.tiles[t2->y*self->map.w+t2->x].f > tentative_f){
						//now t2 is the first node whose f is larger than nl's
						//so t3 is the last node whose f is smaller than nl's
						break;
					}
					t3 = t2;
					t2 = t2->next;
				}
				if(t3 != NULL){
					//the next node to nl is now the node which was next to t3
					nl->next = t3->next;
					//the next node to t3 is now nl
					t3->next = nl;
				}
				else{
					//t3==NULL -> nl goes first
					nl->next = openset_first;
					openset_first = nl;
				}

				tentative_is_better = 1;
			}
			else if(tentative_g < self->map.tiles[nl->y*self->map.w+nl->x].g){
				tentative_is_better = 1;
			}
			if(tentative_is_better){
				int nnum = nl->y*self->map.w+nl->x;
				self->map.tiles[nnum].from = t->y*self->map.w+t->x;
				self->map.tiles[nnum].g = tentative_g;
				self->map.tiles[nnum].h = tentative_h;
				self->map.tiles[nnum].f = tentative_f;
			}
		}
	}

	//Reconstruct the path
	if(path_found){
		//fprintf(stderr, "game_find_path(): path found, reconstructing...\n");
		unsigned int length=0;
		//(n = the goal node = the first node in openset)
		unsigned int i;
		i = t->y*self->map.w+t->x;
		//fprintf(stderr, "first pass, counting length. initial node %i\n", i);
		for(;;){
			length++;
			//fprintf(stderr, "node %i, length %i\n", i, length);
			i = self->map.tiles[i].from;
			if(i == -1) break;
		}
		//fprintf(stderr, "length=%i\n", length);

		path = (struct tpos*)malloc(sizeof(struct tpos)*length);

		unsigned int j=length-1;
		i = t->y*self->map.w+t->x;
		//fprintf(stderr, "second pass, collecting path. initial node %i\n", i);
		for(;;){
			path[j].y = i / self->map.w;
			path[j].x = i - path[j].y * self->map.w;
			//fprintf(stderr, "node %i: path[j=%i]={%i,%i})\n", i, j, path[j].x, path[j].y);
			i = self->map.tiles[i].from;
			if(i == -1) break;
			j--;
		}
		//fprintf(stderr, "game_find_path(): done: ");
		//fprintf(stderr, "game_find_path(): path found: ");
		//for(j=0; j<length; j++) fprintf(stderr, "(%i,%i) ", path[j].x, path[j].y);
		//fprintf(stderr, "\n");
	}
	//else fprintf(stderr, "game_find_path(): path not found!\n");

//if failed, path is NULL and sets are cleared
path_fail:

	t = closedset_first;
	while(t != NULL){
		t2 = t->next;
		free(t);
		t = t2;
	}
	t = openset_first;
	while(t != NULL){
		t2 = t->next;
		free(t);
		t = t2;
	}

	return path;
}

/*unsigned int game_get_number_of_units_going_to(struct game *self, int tx, int ty)
{
	unsigned int i, count=0;
	for(i=0; i<UNIT_LIST_SIZE; i++)
	{
		if(self->units[i].type_id == -1) continue;
		struct tpos p = {-1337,-1337};
		if(unit_going_to(&self->units[i], &p) == -1) continue;
		if(p.x == tx && p.y == ty) count++;
	}
	return count;
}*/


