#ifndef GAME_H
#define GAME_H

#include <SDL/SDL.h>
#include "imglib.h"
#include "map.h"
#include "unit.h"

#define UNIT_LIST_SIZE 300
#define SELECTED_UNITS_SIZE 100

struct game
{
	struct {
		uint8_t reach_target_always_when_other_reached_it_too;
	} settings;
	struct imglib imglib;
	SDL_Surface *screen;
	struct unit units[UNIT_LIST_SIZE];
	int screen_needs_complete_update;
	struct map map;
	SDL_Rect view;
	unsigned int selected_units[SELECTED_UNITS_SIZE];
	unsigned int selected_units_count;
};

int game_co(struct game *self, SDL_Surface *screen);
void game_de(struct game *self);
int game_load(struct game *self);
int game_run(struct game *self);
int game_draw(struct game *self);
//-1 if fail
int game_selection_add(struct game *self, unsigned int i);
void game_selection_clear(struct game *self);

struct tile_info
{
	maptile_flags_t mapflags;
	unsigned int unit_count;
	int unit_i[MAX_UNITS_PER_TILE];
};

int game_get_tile_info(struct game *self, struct tile_info *info, int tx, int ty);
//returns index number to units[] or -1
int game_select_unit_at(struct game *self, int px, int py);

typedef int8_t (*blockedfunc_t)(struct tile_info *);
/*
	Finds a path along the map through tiles.
	blocksfunc should return
		-1: can't go,
		0-127: additional cost in tenths of distance
			(cost = distance * (1 + blocksfunc(info) / 10))
	If path is found, returns a dynamically allocated array of map positions,
	which contains the goal position and the starting position.
	If path is not found, returns NULL
	The caller should free the array after use, if not NULL.
*/
struct tpos * game_find_path(struct game *self, int from_x, int from_y, int to_x, int to_y, blockedfunc_t blockedfunc, struct tpos_linked *additional_blocked);

//unsigned int game_get_number_of_units_going_to(struct game *self, int tx, int ty);

#endif

