#include "imglib.h"
#include "filesys.h"
#include <stdlib.h>

#define FILE_NAME_MAX_LEN 100

int stored_img_co(struct stored_img *self)
{
	self->s = NULL;
	self->name = NULL;
	return 0;
}
void stored_img_de(struct stored_img *self)
{
	if(self->s) SDL_FreeSurface(self->s);
	if(self->name) free(self->name);
}
int stored_img_load_bmp(struct stored_img *self, const char *path, SDL_PixelFormat *screenfmt, Uint32 screenflags, Uint32 colorkey)
{
	SDL_Surface *ts = SDL_LoadBMP(path);
	if(ts == NULL) return -1;
	self->s = SDL_ConvertSurface(ts, screenfmt, screenflags);
	SDL_FreeSurface(ts);
	if(self->s == NULL){
		fprintf(stderr, "stored_img: failed to convert surface\n");
		return -1;
	}
	SDL_SetColorKey(self->s, SDL_SRCCOLORKEY, colorkey);
	return 0;
}

int imglib_co(struct imglib *self)
{
	unsigned int i;
	for(i=0; i<STORED_IMAGES_MAX_COUNT; i++)
	{
		if(stored_img_co(&self->images[i])==-1) return -1;
	}
	self->format_set = 0;
	return 0;
}
void imglib_de(struct imglib *self)
{
	unsigned int i;
	for(i=0; i<STORED_IMAGES_MAX_COUNT; i++)
	{
		stored_img_de(&self->images[i]);
	}
}

void imglib_set_format(struct imglib *self, SDL_PixelFormat *screenfmt, Uint32 screenflags, Uint32 colorkey)
{
	self->screenfmt = screenfmt;
	self->screenflags = screenflags;
	self->colorkey = colorkey;
	self->format_set = 1;
}

int imglib_load(struct imglib *self, const char *path)
{
	if(self->format_set == 0)
	{
		fprintf(stderr, "imglib: format has to be set before loading\n");
		return -1;
	}
	struct dir_listing dir_listing;
	if(dir_listing_co(&dir_listing, path) == -1)
	{
		fprintf(stderr, "imglib: failed to read directory %s\n", path);
		return -1;
	}
	int store_i = 0;
	int type;
	char name[FILE_NAME_MAX_LEN];
	char fullpath[MAX_PATH_LEN+1+FILE_NAME_MAX_LEN];
	while(!dir_listing_get_next(&dir_listing, &type, name, FILE_NAME_MAX_LEN))
	{
		if(strcmp(name, ".")==0) continue;
		if(strcmp(name, "..")==0) continue;
		if(strlen(name) < 5) continue;
		if(strcmp(&name[strlen(name)-4], ".bmp") != 0){
			fprintf(stderr, "imglib: a non-bmp file \"%s\"\n", name);
			continue;
		}
		sprintf(fullpath, "%s/%s", path, name);
		fprintf(stderr, "imglib: loading \"%s\"\n", fullpath);
		if(stored_img_load_bmp(&self->images[store_i], fullpath, self->screenfmt, self->screenflags, self->colorkey) == -1){
			fprintf(stderr, "imglib: failed to load \"%s\"\n", fullpath);
		}
		else{
			int namelen = strlen(name);
			//take out the suffix ".bmp"
			namelen -= 4;
			name[namelen] = 0;
			self->images[store_i].name = (char*)malloc(namelen+1);
			if(self->images[store_i].name != NULL){
				memcpy(self->images[store_i].name, name, namelen+1);
			}
			else{
				fprintf(stderr, "imglib: failed to allocate memory for image name \"%s\"\n", name);
			}
			fprintf(stderr, "imglib: image \"%s\" loaded to index %i\n", name, store_i);
			store_i++;
		}
	}
	if(store_i==0) fprintf(stderr, "imglib: loaded nothing\n");
	dir_listing_de(&dir_listing);
	return 0;
}

//struct stored_img * imglib_get_stored_img(struct imglib *self, const char *name)
SDL_Surface * imglib_get_image(struct imglib *self, const char *name)
{
	fprintf(stderr, "imglib_get_image(self, \"%s\")\n", name);
	unsigned int i;
	for(i=0; i<STORED_IMAGES_MAX_COUNT; i++)
	{
		if(self->images[i].name==NULL) continue;
		if(strcmp(self->images[i].name, name)==0){
			fprintf(stderr, "imglib_get_image: returning pointer %i\n", (int)self->images[i].s);
			return self->images[i].s;
		}
	}
	fprintf(stderr, "imglib_get_image: didn't find, returning NULL\n");
	return NULL;
}

