#include <assert.h>
#include <math.h>
#include "map.h"
#include "common.h"

const char * direction_string(uint8_t d)
{
	switch(d){
	case DIRECTION_NONE: return "DIRECTION_NONE";
	case DIRECTION_R: return "DIRECTION_R";
	case DIRECTION_TR: return "DIRECTION_TR";
	case DIRECTION_T: return "DIRECTION_T";
	case DIRECTION_TL: return "DIRECTION_TL";
	case DIRECTION_L: return "DIRECTION_L";
	case DIRECTION_BL: return "DIRECTION_BL";
	case DIRECTION_B: return "DIRECTION_B";
	case DIRECTION_BR: return "DIRECTION_BR";
	default: return "[INVALID DIRECTION]";
	}
}

uint8_t direction_reverse(uint8_t d)
{
	switch(d){
	case DIRECTION_NONE: return DIRECTION_NONE;
	case DIRECTION_R: return DIRECTION_L;
	case DIRECTION_TR: return DIRECTION_BL;
	case DIRECTION_T: return DIRECTION_B;
	case DIRECTION_TL: return DIRECTION_BR;
	case DIRECTION_L: return DIRECTION_R;
	case DIRECTION_BL: return DIRECTION_TR;
	case DIRECTION_B: return DIRECTION_T;
	case DIRECTION_BR: return DIRECTION_TL;
	default: return DIRECTION_NONE;
	}
}

uint8_t jump_direction(int from_x, int from_y, int to_x, int to_y)
{
	if(abs(to_y - from_y) < 0.001 && abs(to_x - from_x) < 0.001)
		return DIRECTION_NONE;
	float a = atan2f((float)(-(to_y - from_y)), (float)(to_x - from_x));
	if(a < 0.0) a += 2.*PI;
	float c = 1./16. * 2. * PI + 0.001;
	float b;
	b = fabs(a - 0./8. * 2. * PI);
	if(b < c)
		return DIRECTION_R;
	b = fabs(a - 1./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_TR;
	b = fabs(a - 2./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_T;
	b = fabs(a - 3./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_TL;
	b = fabs(a - 4./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_L;
	b = fabs(a - 5./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_BL;
	b = fabs(a - 6./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_B;
	b = fabs(a - 7./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_BR;
	b = fabs(a - 8./8. * 2. * PI);
	c = 1./16. * 2. * PI + 0.001;
	if(b < c)
		return DIRECTION_R;
	fprintf(stderr, "(%i,%i)->(%i,%i) -> a=%f\n", from_x, from_y, to_x, to_y, a);
	assert(0);
}

/*uint8_t jump_direction(int from_x, int from_y, int to_x, int to_y)
{
	if(abs(to_y - from_y) < 0.001 && abs(to_x - from_x) < 0.001)
		return DIRECTION_NONE;
	float a = atan2f((float)(-(to_y - from_y)), (float)(to_x - from_x));
	if(a < 0.0) a += 2.*PI;
	if(fabs(a - 0./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_R;
	if(fabs(a - 1./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_TR;
	if(fabs(a - 2./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_T;
	if(fabs(a - 3./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_TL;
	if(fabs(a - 4./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_L;
	if(fabs(a - 5./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_BL;
	if(fabs(a - 6./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_B;
	if(fabs(a - 7./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_BR;
	if(fabs(a - 8./8. * 2. * PI) < 1./16. * 2. * PI + 0.001)
		return DIRECTION_R;
	fprintf(stderr, "(%i,%i)->(%i,%i) -> a=%f\n", from_x, from_y, to_x, to_y, a);
	assert(0);
}*/

uint8_t tjump_direction(struct tpos *from, struct tpos *to)
{
	return jump_direction(from->x, from->y, to->x, to->y);
}

void tadd_direction(struct tpos *t, uint8_t d)
{
	switch(d){
	case DIRECTION_NONE: break;
	case DIRECTION_R: t->x++; break;
	case DIRECTION_TR: t->x++; t->y--; break;
	case DIRECTION_T: t->y--; break;
	case DIRECTION_TL: t->x--; t->y--; break;
	case DIRECTION_L: t->x--; break;
	case DIRECTION_BL: t->x--; t->y++; break;
	case DIRECTION_B: t->y++; break;
	case DIRECTION_BR: t->x++; t->y++; break;
	default: break;
	}
}

unsigned int tpos_linked_list_count(struct tpos_linked *t)
{
	unsigned int c = 0;
	while(t!=NULL){
		c++;
		t = t->next;
	}
	return c;
}

void tpos_linked_list_add_single_copy(struct tpos_linked **list, struct tpos *a)
{
	struct tpos_linked *l;
	l = (struct tpos_linked*)malloc(sizeof(struct tpos_linked));
	l->x = a->x;
	l->y = a->y;
	l->next = *list;
	*list = l;
}

void tpos_linked_list_clear(struct tpos_linked **list)
{
	struct tpos_linked *t = *list;
	while(t!=NULL){
		struct tpos_linked *t2 = t->next;
		free(t);
		t = t2;
	}
	*list = NULL;
}

float distance(int x1, int y1, int x2, int y2)
{
	int dx = x2-x1;
	int dy = y2-y1;
	return sqrt(dx*dx+dy*dy);
}

/*
Map String:
"[char] [imagename] [flag1] ...\n"
"[char] [imagename] [flag1] ...\n"
"\n[width]\n[height]\n"
"<row1>"
"<row2>"
"..."
Possible flags for map string:
	walkable
	buildable
*/

const char *testmap =
"  bgtest1 walkable buildable\n"
"b bgtest2 walkable\n"
"t trees1\n"
"\n40\n24\n"
"     bb  b  bbbb   bb  ttttttttttttttttt"
"      bb bbbbb    b      b  b      bbttt"
"       bbbbbb  b        b ttttttttbb  tt"
"      bbb b   b  b    bb    tttbbtttt  t"
"b    bbbbbbbttb  bbbb      bttb        t"
"bbb    bbbtttttbbb b      bbtt          "
"bbbb  ttttttttttb   b    bb  tt         "
"bb      bb     bb    bbbb   tt          "
"bbbbtttttttttttttt tttbbbbbbbtt        t"
"bbtttttttttttttttt tttttttb   tt      tt"
"ttt  bb tt         tttttt      ttt  tttt"
"ttt ttbbb  ttttttt ttttttbb             "
"ttt tttttttttttttt   tttttbbb           "
"tt  tttttbbbbtttttt  ttttttbb           "
"tt ttttbbbbbbbbbttttttttttbbb         b "
"tt   bbbbbbbbbbbbttttttttbb          b b"
"tttbbb bb       bbtttbbbb        b   bbb"
"ttbbb   b   bb   bbbb         b   bbbb  "
"t  b      bb                   bbbbbb   "
"tttttttttttttttttttttttttttttbbbbb      "
"t                          ttbbbb       "
"t                           bb          "
"t                          tt           "
"ttttttttttttttttttttttttttttt           "
;

struct identifier
{
	char c;
	SDL_Surface *s;
	maptile_flags_t flags;
};

int map_co(struct map *self)
{
	self->tiles = NULL;
	return 0;
}

void map_de(struct map *self)
{
	map_unload(self);
}

int map_load_string(struct map *self, const char *str, struct imglib *imglib)
{
	unsigned int i, j;
	unsigned int ids_count = 0;
	const char *p = str;
	char tv[MAP_STRING_TEMP_VALUE_SIZE];
	struct identifier ids[MAP_STRING_MAX_IDENTIFIERS]; 
	unsigned int end_of_ids = 0;
	for(i=0; i<MAP_STRING_MAX_IDENTIFIERS; i++){
		ids[i].c = *p;
		ids[i].s = NULL;
		ids[i].flags = 0;
		p+=2;
		j=0;
		do{
			if(j < MAP_STRING_TEMP_VALUE_SIZE-1) tv[j] = *p;
			p++;
			j++;
		} while(*p != ' ' && *p != '\n');
		tv[lower(j, MAP_STRING_TEMP_VALUE_SIZE-1)] = 0;
		ids[i].s = imglib_get_image(imglib, tv);
		if(ids[i].s == NULL){
			fprintf(stderr, "map_load_string: warning: couldn't get image \"%s\"\n", tv);
		}
		else{
			if(ids[i].s->w != MAP_TILE_W || ids[i].s->h != MAP_TILE_H){
				fprintf(stderr, "map_load_string: warning: image \"%s\" w,h (%i,%i) != MAP_TILE_W,H (%i,%i)\n", tv, ids[i].s->w, ids[i].s->h, MAP_TILE_W, MAP_TILE_H);
			}
		}
		fprintf(stderr, "map_load_string: id '%c': image=\"%s\"\n", ids[i].c, tv);
		for(;;){
			if(*p == '\n'){
				p++;
				if(*p == '\n'){
					end_of_ids = 1;
				}
				break;
			}
			p++;
			j=0;
			do{
				if(j < MAP_STRING_TEMP_VALUE_SIZE-1) tv[j] = *p;
				p++;
				j++;
			} while(*p != ' ' && *p != '\n');
			tv[lower(j, MAP_STRING_TEMP_VALUE_SIZE-1)] = 0;
			//parse tv
			if(strcmp(tv, "walkable")==0){
				ids[i].flags |= MAPTILE_WALKABLE;
			}
			else if(strcmp(tv, "buildable")==0){
				ids[i].flags |= MAPTILE_BUILDABLE;
			}
			else{
				fprintf(stderr, "invalid flag: \"%s\"\n", tv);
			}
		}
		fprintf(stderr, "    flags=0x%x\n", ids[i].flags);
		if(end_of_ids){
			break;
		}
	}
	if(i==MAP_STRING_MAX_IDENTIFIERS) ids_count = i;
	else ids_count = i+1;
	j=0;
	do{
		if(j < MAP_STRING_TEMP_VALUE_SIZE-1) tv[j] = *p;
		p++;
		j++;
	} while(*p != '\n');
	tv[lower(j, MAP_STRING_TEMP_VALUE_SIZE-1)] = 0;
	self->w = atoi(tv);
	j=0;
	do{
		if(j < MAP_STRING_TEMP_VALUE_SIZE-1) tv[j] = *p;
		p++;
		j++;
	} while(*p != '\n');
	tv[lower(j, MAP_STRING_TEMP_VALUE_SIZE-1)] = 0;
	self->h = atoi(tv);
	p++;
	int tilecount = self->w*self->h;
	fprintf(stderr, "w=%i, h=%i. allocating %i*sizeof(struct maptile)\n", self->w, self->h, tilecount);
	self->tiles = (struct maptile*)malloc(tilecount*sizeof(struct maptile));
	struct maptile *tp = self->tiles;
	if(tp == NULL){
		fprintf(stderr, "failed to allocate memory for map\n");
		return -1;
	}
	for(i=0; i<tilecount; i++){
		for(j=0; j<ids_count; j++){
			if(*p == ids[j].c) break;
		}
		if(j == ids_count){
			fprintf(stderr, ":couldn't find id '%c'\n", *p);
			assert(0); //DEBUG
			continue;
		}
		fprintf(stderr, "%c", *p);
		tp->flags = ids[j].flags;
		if(tp->flags & MAPTILE_WALKABLE) fprintf(stderr, " ");
		else fprintf(stderr, "X");
		if(i % self->w == 0) fprintf(stderr, "\n");
		tp->s = ids[j].s;
		//fprintf(stderr, ":loaded tile %i: flags=0x%x\n", i, tp->flags);
		tp++;
		p++;
	}
	fprintf(stderr, "\n");
	return 0;
}
void map_unload(struct map *self)
{
	if(self->tiles != NULL) free(self->tiles);
	self->tiles = NULL;
	self->w = 0;
	self->h = 0;
}
void map_draw(struct map *self, SDL_Surface *surface, int map_x, int map_y, int dst_x, int dst_y, int map_w, int map_h)
{
	assert(self->tiles != NULL);
	//fprintf(stderr, "map_draw: map=%i,%i, dst=%i,%i, size=%i,%i)\n", map_x, map_y, dst_x, dst_y, map_w, map_h);
	if(map_x >= self->w*MAP_TILE_W || map_y >= self->h*MAP_TILE_H) return;
	if(map_x < 0){ map_w += map_x; dst_x -= map_x; map_x = 0; }
	if(map_y < 0){ map_h += map_y; dst_y -= map_y; map_y = 0; }
	if(map_w <= 0 || map_h <= 0) return;
	if(map_x + map_w > self->w*MAP_TILE_W) map_w = self->w*MAP_TILE_W - map_x;
	if(map_y + map_h > self->h*MAP_TILE_H) map_h = self->h*MAP_TILE_H - map_y;
	int tx = map_x / MAP_TILE_W;
	int ty = map_y / MAP_TILE_H;
	int tx2 = (map_x+map_w-1) / MAP_TILE_W;
	int ty2 = (map_y+map_h-1) / MAP_TILE_H;
	int iy, ix;
	int cy, cx;
	SDL_Rect src, dst;

	cy=0;
	for(iy=ty; iy<=ty2; iy++){
		if(iy==ty) src.y = map_y - ty * MAP_TILE_H;
		else       src.y = 0;
		if(iy!=ty2) src.h = lower(MAP_TILE_H - src.y, map_h);
		else        src.h = lower(map_y + map_h - ty2 * MAP_TILE_H, map_h);
		cx=0;
		for(ix=tx; ix<=tx2; ix++){
			if(ix==tx) src.x = map_x - tx * MAP_TILE_W;
			else       src.x = 0;
			if(ix!=tx2) src.w = lower(MAP_TILE_W - src.x, map_w);
			else        src.w = lower(map_x + map_w - tx2 * MAP_TILE_W, map_w);
			dst.x = dst_x+cx;
			dst.y = dst_y+cy;
			//fprintf(stderr, "blitting tile (%i,%i) (src=%i,%i, size=%i,%i, dst=%i,%i)\n", ix, iy, src.x, src.y, src.w, src.h, dst.x, dst.y);
			SDL_BlitSurface(self->tiles[iy*self->w+ix].s, &src, surface, &dst);
			cx += src.w;
		}
		cy += src.h;
	}
}


