#ifndef MAP_H
#define MAP_H

#include <stdint.h>
#include <SDL/SDL.h>
#include "imglib.h"

//#define MAP_STRING_IMAGE_NAME_SIZE 20
#define MAP_STRING_MAX_IDENTIFIERS 20
#define MAP_STRING_TEMP_VALUE_SIZE 20

#define MAPTILE_WALKABLE 0x01
#define MAPTILE_BUILDABLE 0x02

enum{
	DIRECTION_R,
	DIRECTION_TR,
	DIRECTION_T,
	DIRECTION_TL,
	DIRECTION_L,
	DIRECTION_BL,
	DIRECTION_B,
	DIRECTION_BR,
	DIRECTION_NONE
};

const char * direction_string(uint8_t d);

uint8_t direction_reverse(uint8_t d);

//tile position
struct tpos
{
	int16_t x, y;
};
struct tpos_linked
{
	int16_t x, y;
	struct tpos_linked *next;
};

uint8_t jump_direction(int from_x, int from_y, int to_x, int to_y);
uint8_t tjump_direction(struct tpos *from, struct tpos *to);
void tadd_direction(struct tpos *t, uint8_t d);

unsigned int tpos_linked_list_count(struct tpos_linked *t);
void tpos_linked_list_add_single_copy(struct tpos_linked **list, struct tpos *a);
void tpos_linked_list_clear(struct tpos_linked **list);

//pixel position
struct ppos
{
	int16_t x, y;
};

float distance(int x1, int y1, int x2, int y2);

typedef uint8_t maptile_flags_t;

struct maptile
{
	maptile_flags_t flags;
	SDL_Surface *s;
	//for pathfinding:
	/*
		-2 = not calculated yet
		-1 = can't go
		0 = can go
		1-127 = additional cost in tenths of distance in tiles
	*/
	int8_t blocked;
	float g;
	float h;
	float f;
	int16_t from; //-1 = not set
};

struct map
{
	struct maptile *tiles;
	int w, h;
};

const char *testmap;

int map_co(struct map *self);
void map_de(struct map *self);
int map_load_string(struct map *self, const char *str, struct imglib *imglib);
void map_unload(struct map *self);
void map_draw(struct map *self, SDL_Surface *surface, int map_x, int map_y, int dst_x, int dst_y, int map_w, int map_h);

#endif

